#' Creates POIs for a given data theme
#' @param srcData  (data.frame) (data frame) DF with two columns:
#             1 - COMID
#             2 - Unique ID value for POI (Streamgage ID, etc.)
#' @param nhdDF (sf data.frame) valid data frame of NHD flowlines
#' @param IDfield character) Name of 'Type' field to be modified in POI
#'
#' @return (sf data.frame) OIs for the specific data theme
#' @importFrom nhdplusTools get_node
#' @importFrom dplyr filter mutate inner_join select
#' @export
POI_creation<-function(srcData, nhdDF, IDfield){

  # Give generic ID to Identity field
  colnames(srcData) <- c("COMID", "ID")

  sub_segs <- filter(nhdDF, COMID %in% srcData$COMID)

  POIs <- sub_segs %>%
    get_node(., position = "end") %>%
    mutate(COMID = sub_segs$COMID) %>%
    inner_join(srcData %>% select(COMID, ID), by = "COMID") %>%
    mutate(!!(paste0("Type_", IDfield)) := ID) %>%
    select(COMID, !!(paste0("Type_", IDfield))) %>%
    st_as_sf()

  return(POIs)
}


#' Adds the type attribute for co-located POIs of multiple themes
#' @param new_POIs (sf data.frame) new POIs to be tested against existing
#' @param POIs  (sf data.frame) Existing POIs
#' @param IDfield (character) Name of 'Type' field to be modified in existing POI
#' @param nexus (logical) whether to consider existing nexus/event locations when adding to POI information
#' @param bind (logical) whether to bind non co-located POIs to data frame or just
#'             attribute existing POIs
#'
#' @return (sf data.frame) xisting POIs with Type field modified for
#'            duplicate POIs for given data theme
#' @export
#' @importFrom nhdplusTools st_compatibalize
#' @importFrom dplyr rename mutate left_join select
#' @importFrom sf st_drop_geometry
#'
addType <- function(new_POIs, POIs, IDfield, nexus = TRUE, bind = TRUE){

  new_POIs <- st_compatibalize(new_POIs, POIs)

  # subset Nexus POIs from Type
  if(nexus){
    nexus_POIs <- filter(POIs, nexus == TRUE)
    POIs <- filter(POIs, nexus %in% c(FALSE, NA))
  }

  if(paste0("Type_", IDfield) %in% colnames(POIs)){
    POIs_exist <- POIs %>%
      rename(ID = !!(paste0("Type_", IDfield)))
  } else {
    POIs_exist <- POIs %>%
      mutate(ID = NA)
  }

  # Add columns missing between master POI and new set
  missing_cols <- colnames(POIs)[!colnames(POIs) %in% colnames(new_POIs)]
  for(col in missing_cols){
    new_POIs <-  new_POIs %>%
      mutate(!!col := NA)
  }

  POIs_newAtt <- filter(new_POIs, COMID %in% POIs$COMID) %>%
    rename(ID2 = !!(paste0("Type_", IDfield)))

  POIs_fin <- POIs_exist %>%
    left_join(st_drop_geometry(POIs_newAtt) %>%
                select(COMID, ID2), by = "COMID") %>% #, all.x = TRUE) %>%
    mutate(ID = ifelse(!is.na(ID2), ID2, ID)) %>%
    rename(!!(paste0("Type_", IDfield)) := ID) %>%
    select(-ID2)

  # Bind unless indicated not to
  if(bind){
    POIs_fin <- data.table::rbindlist(list(POIs_fin, filter(new_POIs, !COMID %in% POIs_fin$COMID)),
                                      fill = TRUE) %>%
      st_as_sf()
  }

  # Add nexus back in if excluded
  if(nexus){
    POIs_fin <- data.table::rbindlist(list(POIs_fin, nexus_POIs), fill = TRUE) %>%
      st_as_sf()
  }

  return(POIs_fin)
}
