#' Creates POIs for a given data theme
#' @param srcData  (data.frame) (data frame) DF with two columns:
#             1 - COMID
#             2 - Unique ID value for POI (Streamgage ID, etc.)
#' @param nhdDF (sf data.frame) valid data frame of NHD flowlines
#' @param IDfield character) Name of 'Type' field to be modified in POI
#'
#' @return (sf data.frame) OIs for the specific data theme
#' @importFrom nhdplusTools get_node
#' @importFrom dplyr filter mutate inner_join select
#' @export
POI_creation<-function(srcData, nhdDF, IDfield){

  # Give generic ID to Identity field
  colnames(srcData) <- c("COMID", "ID")

  sub_segs <- filter(nhdDF, COMID %in% srcData$COMID)

  POIs <- sub_segs %>%
    get_node(., position = "end") %>%
    mutate(COMID = sub_segs$COMID) %>%
    inner_join(srcData %>% select(COMID, ID), by = "COMID") %>%
    mutate(!!(paste0("Type_", IDfield)) := ID) %>%
    select(COMID, !!(paste0("Type_", IDfield))) %>%
    st_as_sf()

  return(POIs)
}


#' Adds the type attribute for co-located POIs of multiple themes
#' @param new_POIs (sf data.frame) new POIs to be tested against existing
#' @param POIs  (sf data.frame) Existing POIs
#' @param IDfield (character) Name of 'Type' field to be modified in existing POI
#' @param nexus (logical) whether to consider existing nexus/event locations when adding to POI information
#' @param bind (logical) whether to bind non co-located POIs to data frame or just
#'             attribute existing POIs
#'
#' @return (sf data.frame) xisting POIs with Type field modified for
#'            duplicate POIs for given data theme
#' @export
#' @importFrom nhdplusTools st_compatibalize
#' @importFrom dplyr rename mutate left_join select
#' @importFrom sf st_drop_geometry
#'
addType <- function(new_POIs, POIs, IDfield, nexus = TRUE, bind = TRUE){

  # Compatibalize new points to add
  new_POIs <- st_compatibalize(new_POIs, POIs)

  # If nexus not a field, add it to existing POIs
  if(!"nexus" %in% colnames(POIs)){
    POIs <- mutate(POIs, nexus = FALSE)
  }

  if(!"nexus" %in% colnames(new_POIs)){
    new_POIs <- mutate(new_POIs, nexus = FALSE)
  }

  # Check if ID field exists in POIs, add if not
  if(paste0("Type_", IDfield) %in% colnames(POIs)){
    POIs_exist <- POIs %>%
      rename(ID = !!(paste0("Type_", IDfield)))
  } else {
    POIs_exist <- POIs %>%
      mutate(ID = NA)
  }

  # subset Nexus POIs from old POIs (if they exist)
  if(nexus){
    # subset existing nexus/split locations to keep unique
    nexus_POIs <- filter(POIs, nexus)
    # Non-nexus POIs
    POIs <- filter(POIs, !nexus)
  }

  # Non-nexus POIs to bind attributes to existing POI
  POIs_newAtt <- filter(new_POIs, COMID %in% POIs$COMID,
                        !nexus) %>%
    rename(ID2 = !!(paste0("Type_", IDfield)))

  # Bind attributes
  POIs_orig <- POIs_exist %>%
    left_join(st_drop_geometry(POIs_newAtt) %>%
                select(COMID, ID2), by = "COMID") %>%
    mutate(ID = ifelse(!is.na(ID2), ID2, ID)) %>%
    rename(!!(paste0("Type_", IDfield)) := ID) %>%
    select(-ID2)

  # Non-nex POI new to POI set
  POIs_new <- filter(new_POIs, !COMID %in% POIs$COMID)

  # If binding and non-nexus HL values exist to bind
  if(bind){
    # If POIs exist for new HL reference
    if (nrow(POIs_new) > 0){
      POIs_fin <- data.table::rbindlist(list(POIs_orig,
                                             POIs_new),
                                        fill = TRUE) %>%
        st_as_sf()
    } else {
      POIs_fin <- POIs_orig
    }
  } else {
    POIs_fin <- POIs_orig
  }

  # Split events that are added
  new_nexus <- filter(new_POIs, nexus)

  # Bind and is nexus
  if(nrow(new_nexus) > 0){
    POIs_fin <- data.table::rbindlist(list(POIs_orig, new_nexus),
                                      fill = TRUE) %>%
      st_as_sf()
  }

  return(POIs_fin)
}
